# -*- coding: utf-8 -*-
"""
Created on Sat Feb 17 15:57:18 2018

@author: Callie Darsey

Given  the resized color frame, its flow, and exemplar(s) information for the
frame, form the three versions of  keep (1) / toss (0) components

component_exploration_04
"""



import numpy as np
import cv2
import example_of_DFT_01 as dft_hist_fns


def force_neg_pi_to_pi(value):
    '''
    This function takes the value and adds or subtracts pi until a value in
    [-pi, pi] is reached and returned.
    '''
    while value > np.pi or value < -np.pi:
        
        if value > np.pi:
            
            value = value - 2 * np.pi
            
        elif value < -np.pi:
            
            value = value + 2 * np.pi
            
    return value
    

def get_flow_mag_in_phase_bins(flow, mask, start, bin_size):
    '''
    Return a histogram with flow phase bins [start, start + 1 * bin_size),
    [start + 2 * bin_size), ..., [start + n * size, start) adjusted such that
    the bins capture angles around the unit circle taking values [-pi, pi].
    
    For example, if start = -np.pi / 8 and bin_size = np.pi / 4, the bins are
    
    [    -pi / 8,      pi / 8),
    [     pi / 8,  3 * pi / 8),
    [ 3 * pi / 8,  5 * pi / 8),
    [ 5 * pi / 8,  7 * pi / 8),
    [ 7 * pi / 8, -7 * pi / 8),
    [-7 * pi / 8, -5 * pi / 8),
    [-5 * pi / 8, -3 * pi / 8),
    [-3 * pi / 8,     -pi / 8)
    
    Three versions of this histogram are returned.
    
    1) The value in each bin will be the sum of the flow magnitude values of
       pixels whose flow phase was within the phase bin range.
    
    2) The histogram described in 1) with each value divided by the phase bin
       count (if the count is nonzero)
       
    3) The histogram described in 1) with each value divided by the mask (i.e.
        component) area

    Inputs:
        
    flow - two channel array holding the bin sizes
    mask - mask of the connected component
    start - angle in radians to start the phase bins at
    bin_size - angle in radians of the bin widths

    '''
    assert bin_size <= 2 * np.pi and bin_size > 0
    
    # force start to be in [-pi, pi]
    start = force_neg_pi_to_pi(start)
    
    # set up the bin edges
    bin_edges = [start]
    
    angles_travelled = 0
    while True:
         
        angles_travelled = angles_travelled + bin_size
        
        if angles_travelled >= 2 * np.pi:
            
            bin_edges.append(start)
            
            break
        
        else:

            new_edge = force_neg_pi_to_pi(bin_edges[-1] + bin_size)
            
            bin_edges.append(new_edge)

    # get the area of the mask (i.e. the area of the component)
    area = np.count_nonzero(mask)

    # get the flow into magnitude and angle form
    flow_mag = cv2.magnitude(flow[:, :, 0], flow[:, :, 1])
    flow_phase = cv2.phase(flow[:, :, 0], flow[:, :, 1])

    # find the sums and counts for the bins
    
    bin_sums = np.zeros((1, len(bin_edges)-1), dtype=np.float32)
    bin_counts = np.zeros((1, len(bin_edges)-1), dtype=np.float32)
    
    r_ind, c_ind = np.where(mask)
    
    for r, c in zip(r_ind, c_ind):
        
        flow_phase_value = force_neg_pi_to_pi(flow_phase[r, c])      
        
        for  i, edge_pair in enumerate(zip(bin_edges[0:-1], bin_edges[1:])):
            
            edge_inclusive, edge_exclusive = edge_pair
            
            # The bin does not cross pi on unit circle
            if edge_inclusive < edge_exclusive:
                
                if edge_inclusive <= flow_phase_value and \
                                     flow_phase_value < edge_exclusive:
                                       
                    bin_sums[0, i] = bin_sums[0, i] + flow_mag[r, c]
                    
                    bin_counts[0, i] = bin_counts[0, i] + 1
                    
                    break
                    
            # The bin crosses pi on the unit circle
            elif edge_inclusive > edge_exclusive:
                
                if edge_inclusive <= flow_phase_value and \
                                     flow_phase_value <= np.pi or \
                           -np.pi <= flow_phase_value and \
                                     flow_phase_value < edge_exclusive:
                
                    bin_sums[0, i] = bin_sums[0, i] + flow_mag[r, c]
                    
                    bin_counts[0, i] = bin_counts[0, i] + 1
                    
                    break
    
    # divide sums by bin counts            
    bin_sums_div_by_count = np.copy(bin_sums)

    bin_sums_div_by_count[bin_counts != 0] = \
        bin_sums_div_by_count[bin_counts != 0] / bin_counts[bin_counts != 0]
        
    # divide sums by area
    bin_sums_div_by_area = bin_sums / area
    
    return bin_sums, bin_sums_div_by_count, bin_sums_div_by_area
        

# column indices into features
FEATURE_B_HIST_LEFT = 0
FEATURE_B_HIST_RIGHT = 255
FEATURE_G_HIST_LEFT = 256
FEATURE_G_HIST_RIGHT = 511
FEATURE_R_HIST_LEFT = 512
FEATURE_R_HIST_RIGHT = 767

FEATURE_H_HIST_LEFT = 768
FEATURE_H_HIST_RIGHT = 947
FEATURE_S_HIST_LEFT = 948
FEATURE_S_HIST_RIGHT = 1203
FEATURE_V_HIST_LEFT = 1204
FEATURE_V_HIST_RIGHT = 1459

FEATURE_Y_HIST_LEFT = 1460
FEATURE_Y_HIST_RIGHT = 1715
FEATURE_Cr_HIST_LEFT = 1716
FEATURE_Cr_HIST_RIGHT = 1971
FEATURE_Cb_HIST_LEFT = 1972
FEATURE_Cb_HIST_RIGHT = 2227

FEATURE_B_PDF_LEFT = 2228
FEATURE_B_PDF_RIGHT = 2483
FEATURE_G_PDF_LEFT = 2484
FEATURE_G_PDF_RIGHT = 2739
FEATURE_R_PDF_LEFT = 2740
FEATURE_R_PDF_RIGHT = 2995

FEATURE_H_PDF_LEFT = 2996
FEATURE_H_PDF_RIGHT = 3175
FEATURE_S_PDF_LEFT = 3176
FEATURE_S_PDF_RIGHT = 3431
FEATURE_V_PDF_LEFT = 3432
FEATURE_V_PDF_RIGHT = 3687

FEATURE_Y_PDF_LEFT = 3688
FEATURE_Y_PDF_RIGHT = 3943
FEATURE_Cr_PDF_LEFT = 3944
FEATURE_Cr_PDF_RIGHT = 4199
FEATURE_Cb_PDF_LEFT = 4200
FEATURE_Cb_PDF_RIGHT = 4455

FEATURE_CANNY_EDGE_TO_AREA_RATIO = 4456

FEATURE_DFT_MAG_SPEC_HIST_LEFT = 4457
FEATURE_DFT_MAG_SPEC_HIST_RIGHT = 4520

FEATURE_DFT_PHASE_SPEC_HIST_LEFT = 4521
FEATURE_DFT_PHASE_SPEC_HIST_RIGHT = 4584

FEATURE_FLOW_PHASE_BIN_MAG_SUM_HIST_LEFT = 4585
FEATURE_FLOW_PHASE_BIN_MAG_SUM_HIST_RIGHT = 4592

FEATURE_FLOW_PHASE_BIN_MAG_SUM_HIST_DIV_COUNT_LEFT = 4593
FEATURE_FLOW_PHASE_BIN_MAG_SUM_HIST_DIV_COUNT_RIGHT = 4600

FEATURE_FLOW_PHASE_BIN_MAG_SUM_HIST_DIV_AREA_LEFT = 4601
FEATURE_FLOW_PHASE_BIN_MAG_SUM_HIST_DIV_AREA_RIGHT = 4608

FEATURE_AREA = 4609
FEATURE_PERIMETER = 4610

FEATURE_MIN_AREA_RECT_ORIENT = 4611
FEATURE_AREA_TO_MIN_AREA_RECT = 4612
FEATURE_MIN_AREA_RECT_SHORT_TO_LONG_SIDES_RATIO = 4613

features_vector_size = FEATURE_MIN_AREA_RECT_SHORT_TO_LONG_SIDES_RATIO + 1

def get_features(cc, frame, flow):
    '''
    cc - Output of cv2.connectedComponentsWithStats() on components with
         appended element holding the label 1 for 'keep' and 0 for 'toss' and
         -1 for the background component with label 0.         
    frame - resized color frame
    flow - two channel flow for resized color frame
    
    Return cc with another element holding the features vector. (The
    background is filled with -1 values.)
    '''
   
    # append another element to hold features (fill background with -1)
    cc.append([])
    cc[-1].append(-1 * np.ones((1, features_vector_size), dtype=np.float32))
    
    # convert the color array to YCrCb and HSV
    frame_HSV = cv2.cvtColor(frame, cv2.COLOR_BGR2HSV)
    frame_YCrCb = cv2.cvtColor(frame, cv2.COLOR_BGR2YCrCb)
    
    
    # BGRHSVYCrCb
    combine = [frame, frame_HSV, frame_YCrCb]
    BGRHSVYCrCb = np.stack(combine, axis=2)
    BGRHSVYCrCb = np.reshape(BGRHSVYCrCb,(frame.shape[0], frame.shape[1],
                                          frame.shape[2] * len(combine)))
    
    # prep to go through the channels of BGRHSVYCrCb and get histograms
    
    # bin edges
    bin_edges_0to255 = np.arange(0, 257, 1)
    bin_edges_0to180 = np.arange(0, 181, 1)
    bin_edges_arrays = [bin_edges_0to255 for i in range(3)] + \
                       [bin_edges_0to180] + [bin_edges_0to255 for i in range(5)]
                
    # indices: start inclusive, end exclusive
    start_end_pairs_hist = [[FEATURE_B_HIST_LEFT, FEATURE_B_HIST_RIGHT + 1],
                            [FEATURE_G_HIST_LEFT, FEATURE_G_HIST_RIGHT + 1],
                            [FEATURE_R_HIST_LEFT, FEATURE_R_HIST_RIGHT + 1],
                       
                            [FEATURE_H_HIST_LEFT, FEATURE_H_HIST_RIGHT + 1],
                            [FEATURE_S_HIST_LEFT, FEATURE_S_HIST_RIGHT + 1],
                            [FEATURE_V_HIST_LEFT, FEATURE_V_HIST_RIGHT + 1],
                       
                            [FEATURE_Y_HIST_LEFT, FEATURE_Y_HIST_RIGHT + 1],
                            [FEATURE_Cr_HIST_LEFT, FEATURE_Cr_HIST_RIGHT + 1],
                            [FEATURE_Cb_HIST_LEFT, FEATURE_Cb_HIST_RIGHT + 1]]
                       
    start_end_pairs_pdf = [[FEATURE_B_PDF_LEFT, FEATURE_B_PDF_RIGHT + 1],
                           [FEATURE_G_PDF_LEFT, FEATURE_G_PDF_RIGHT + 1],
                           [FEATURE_R_PDF_LEFT, FEATURE_R_PDF_RIGHT + 1],
                       
                           [FEATURE_H_PDF_LEFT, FEATURE_H_PDF_RIGHT + 1],
                           [FEATURE_S_PDF_LEFT, FEATURE_S_PDF_RIGHT + 1],
                           [FEATURE_V_PDF_LEFT, FEATURE_V_PDF_RIGHT + 1],
                       
                           [FEATURE_Y_PDF_LEFT, FEATURE_Y_PDF_RIGHT + 1],
                           [FEATURE_Cr_PDF_LEFT, FEATURE_Cr_PDF_RIGHT + 1],
                           [FEATURE_Cb_PDF_LEFT, FEATURE_Cb_PDF_RIGHT + 1]]
    
    # canny
    '''https://docs.opencv.org/3.1.0/da/d22/tutorial_py_canny.html'''
    minVal = 50
    maxVal = 150
    edges = cv2.Canny(frame_YCrCb[:, :, 0], minVal, maxVal)
    
    
    
    # now loop through the rest of the components and gather features
    for label in np.unique(cc[1])[1:]:
    
        features_vector = np.zeros((1, features_vector_size), dtype=np.float32)
            
        cc_bool_mask = (cc[1] == label)
        
        dilated_cc_mask = cv2.dilate(cc_bool_mask.astype(np.uint8),
                                     kernel=np.ones((3,3)))
        
        cc_dilated_cc = cv2.connectedComponentsWithStats(dilated_cc_mask)
        cc_dilated_bool_mask = cc_dilated_cc[1].astype(np.bool)
        cc_dilated_mask = cc_dilated_bool_mask.astype(np.uint8)
        
        # form the color space histograms    
        for i, info in enumerate(zip(bin_edges_arrays, start_end_pairs_hist)):
            
            bin_edges, start_end = info
    
            hist, bin_edges = np.histogram(BGRHSVYCrCb[:, :, i] \
                                           [cc_dilated_bool_mask],
                                           bin_edges)
        
            features_vector[:, start_end[0]:start_end[1]] = hist

        # form the color space PDFs    
        for i, info in enumerate(zip(bin_edges_arrays, start_end_pairs_pdf)):
            
            bin_edges, start_end = info
    
            hist, bin_edges = np.histogram(BGRHSVYCrCb[:, :, i] \
                                           [cc_dilated_bool_mask],
                                           bin_edges)
        
            features_vector[:, start_end[0]:start_end[1]] = hist / np.sum(hist)
               
            
        # get the canny edge to area ratio (using the dilated component area)
        print('\n', type(edges), edges.dtype, edges.shape, '\n',
              type(cc_dilated_mask), cc_dilated_mask.dtype,
              cc_dilated_mask.shape, '\n')
        edges_in_dilated_cc = cv2.bitwise_and(edges, edges, mask=cc_dilated_mask)
        
        features_vector[:, FEATURE_CANNY_EDGE_TO_AREA_RATIO] = \
            np.count_nonzero(edges_in_dilated_cc) / \
            np.count_nonzero(cc_dilated_mask)
    
        # discrete Fourier transform (DFT) magnitude and phase spectrum histograms
        
        # get the grayscale image crop with the component
        crop = frame_YCrCb[:, :, 0][cc_dilated_cc[2][1, cv2.CC_STAT_TOP] : \
                                    cc_dilated_cc[2][1, cv2.CC_STAT_TOP] + \
                                    cc_dilated_cc[2][1, cv2.CC_STAT_HEIGHT], 
                                    cc_dilated_cc[2][1, cv2.CC_STAT_LEFT] : \
                                    cc_dilated_cc[2][1, cv2.CC_STAT_LEFT] + \
                                    cc_dilated_cc[2][1, cv2.CC_STAT_WIDTH]]
     
        mag_spec_hist, phase_spec_hist, _, _, _ = \
            dft_hist_fns.get_radians_per_pixel_histograms(crop)
            
        features_vector[:, FEATURE_DFT_MAG_SPEC_HIST_LEFT : \
                           FEATURE_DFT_MAG_SPEC_HIST_RIGHT + 1] = \
                        np.reshape(mag_spec_hist, (1, -1))
                 
                        
        features_vector[:, FEATURE_DFT_PHASE_SPEC_HIST_LEFT : \
                           FEATURE_DFT_PHASE_SPEC_HIST_RIGHT + 1] = \
                        np.reshape(phase_spec_hist, (1, -1))
    
            
        # flow phase bins to flow magnitude sums 'histogram'; using dilated mask
        bin_sums, bin_sums_div_by_count, bin_sums_div_by_area = \
            get_flow_mag_in_phase_bins(flow, cc_dilated_mask,
                                       -np.pi / 8, np.pi / 4)
            
        features_vector[:, FEATURE_FLOW_PHASE_BIN_MAG_SUM_HIST_LEFT : \
                           FEATURE_FLOW_PHASE_BIN_MAG_SUM_HIST_RIGHT + 1] = \
            bin_sums
        
        features_vector[:, FEATURE_FLOW_PHASE_BIN_MAG_SUM_HIST_DIV_COUNT_LEFT : \
                           FEATURE_FLOW_PHASE_BIN_MAG_SUM_HIST_DIV_COUNT_RIGHT \
                           + 1] = bin_sums_div_by_count
                        
        features_vector[:, FEATURE_FLOW_PHASE_BIN_MAG_SUM_HIST_DIV_AREA_LEFT : \
                           FEATURE_FLOW_PHASE_BIN_MAG_SUM_HIST_DIV_AREA_RIGHT \
                           + 1] = bin_sums_div_by_area
                        
        # area
        area = np.count_nonzero(cc_dilated_mask)
        features_vector[:, FEATURE_AREA] = area
        
        # feature perimeter
        kernel = np.ones((3, 3))
        
        features_vector[:, FEATURE_PERIMETER] = \
            np.count_nonzero(cv2.subtract(cv2.dilate(cc_dilated_mask, kernel),
                                          cv2.erode(cc_dilated_mask, kernel)))
    
        # minimum area rectangle features
        
        # get the external contour points 
        contours = cv2.findContours(cc_dilated_mask, cv2.RETR_EXTERNAL,
                                    cv2.CHAIN_APPROX_SIMPLE)
        
        # reshape the points into N x 2 for passing to cv2.minAreaRect()
        contour_points = contours[1][0].reshape((contours[1][0].shape[0], -1))
        
        # get the minimum area rectangle info
        rect_center_point, rect_shape, rotation = cv2.minAreaRect(contour_points)
        
        # separate the rotations
        width_along_rotated_x, height_along_rotated_y = rect_shape
        
        '''when width_along_rotated_x is smaller than height_along_rotated_y, 
        adjust the rotation to distinguish orientation'''
        if width_along_rotated_x < height_along_rotated_y:
            
            if abs(rotation) != 90:
                
                rotation = -rotation
                
            else:
                
                rotation = 0
                
        #-force -90 to become 90'''            
        elif rotation == -90:
            
            rotation = 90
    
        # save the orientation feature
        features_vector[:, FEATURE_MIN_AREA_RECT_ORIENT] = rotation
    
        # save the ratio of component area to minimum area rectangle
        features_vector[:, FEATURE_AREA_TO_MIN_AREA_RECT] = \
            area / (width_along_rotated_x * height_along_rotated_y)
            
        # save ratio of short side to long side of minimum area rectangle
        features_vector[:, FEATURE_MIN_AREA_RECT_SHORT_TO_LONG_SIDES_RATIO] = \
            min(rect_shape) / max(rect_shape)
            
        if min(rect_shape) < 0 or max(rect_shape) < 0:
            print(rect_shape)
            
        # save the features vector for the label
        cc[-1].append(features_vector)
        
    return cc


    


























