# -*- coding: utf-8 -*-
"""
Created on Mon May  7 15:43:32 2018

@author: Callie Darsey

collect features from predominately_head and predominately_hand

features_from_reference_bin_size_8_wheeler_01
"""

import cv2
import numpy as np
import pickle
import matplotlib.pyplot as plt
import os


def locate_window(title):
    '''
    This function places the window for openCV
    
    '''
    
    move_window = [0, -30]
    cv2.namedWindow(title, cv2.WINDOW_GUI_NORMAL | cv2.WINDOW_AUTOSIZE)
    cv2.moveWindow(title, move_window[0], move_window[1])



def force_neg_pi_to_pi(value):
    '''
    This function takes the value and adds or subtracts pi until a value in
    [-pi, pi] is reached and returned.
    '''
    while value > np.pi or value < -np.pi:
        
        if value > np.pi:
            
            value = value - 2 * np.pi
            
        elif value < -np.pi:
            
            value = value + 2 * np.pi
            
    return value


    
def get_flow_mag_in_phase_bins(flow_mag, flow_phase, mask, start, bin_size):
    '''
    Return a histogram with flow phase bins [start, start + 1 * bin_size),
    [start + 2 * bin_size), ..., [start + n * size, start) adjusted such that
    the bins capture angles around the unit circle taking values [-pi, pi].
    
    For example, if start = -np.pi / 8 and bin_size = np.pi / 4, the bins are
    
    [    -pi / 8,      pi / 8),
    [     pi / 8,  3 * pi / 8),
    [ 3 * pi / 8,  5 * pi / 8),
    [ 5 * pi / 8,  7 * pi / 8),
    [ 7 * pi / 8, -7 * pi / 8),
    [-7 * pi / 8, -5 * pi / 8),
    [-5 * pi / 8, -3 * pi / 8),
    [-3 * pi / 8,     -pi / 8)
    
    Three versions of this histogram are returned.
    
    1) The value in each bin will be the sum of the flow magnitude values of
       pixels whose flow phase was within the phase bin range.
    
    2) The histogram described in 1) with each value divided by the phase bin
       count (if the count is nonzero)
       
    3) The histogram described in 1) with each value divided by the mask (i.e.
        component) area

    Inputs:
        
    flow_mag -  magnitude of the flow
    flow_phase - angle of the flow in radians
    mask - mask of the connected component
    start - angle in radians to start the phase bins at
    bin_size - angle in radians of the bin widths

    '''
    assert bin_size <= 2 * np.pi and bin_size > 0
    
    # force start to be in [-pi, pi]
    start = force_neg_pi_to_pi(start)
    
    # set up the bin edges
    bin_edges = [start]
    
    angles_travelled = 0
    while True:
         
        angles_travelled = angles_travelled + bin_size
        
        if angles_travelled >= 2 * np.pi:
            
            bin_edges.append(start)
            
            break
        
        else:

            new_edge = force_neg_pi_to_pi(bin_edges[-1] + bin_size)
            
            bin_edges.append(new_edge)

    # get the area of the mask (i.e. the area of the component)
    area = np.count_nonzero(mask)


    # find the sums and counts for the bins
    
    bin_sums = np.zeros((1, len(bin_edges)-1), dtype=np.float32)
    bin_counts = np.zeros((1, len(bin_edges)-1), dtype=np.float32)
    
    r_ind, c_ind = np.where(mask)
    
    for r, c in zip(r_ind, c_ind):
        
        flow_phase_value = force_neg_pi_to_pi(flow_phase[r, c])      
        
        for  i, edge_pair in enumerate(zip(bin_edges[0:-1], bin_edges[1:])):
            
            edge_inclusive, edge_exclusive = edge_pair
            
            # The bin does not cross pi on unit circle
            if edge_inclusive < edge_exclusive:
                
                if edge_inclusive <= flow_phase_value and \
                                     flow_phase_value < edge_exclusive:
                                       
                    bin_sums[0, i] = bin_sums[0, i] + flow_mag[r, c]
                    
                    bin_counts[0, i] = bin_counts[0, i] + 1
                    
                    break
                    
            # The bin crosses pi on the unit circle
            elif edge_inclusive > edge_exclusive:
                
                if edge_inclusive <= flow_phase_value and \
                                     flow_phase_value <= np.pi or \
                           -np.pi <= flow_phase_value and \
                                     flow_phase_value < edge_exclusive:
                
                    bin_sums[0, i] = bin_sums[0, i] + flow_mag[r, c]
                    
                    bin_counts[0, i] = bin_counts[0, i] + 1
                    
                    break
    
    # divide sums by bin counts            
    bin_sums_div_by_count = np.copy(bin_sums)

    bin_sums_div_by_count[bin_counts != 0] = \
        bin_sums_div_by_count[bin_counts != 0] / bin_counts[bin_counts != 0]
        
    # divide sums by area
    bin_sums_div_by_area = bin_sums / area
    
    return bin_sums, bin_sums_div_by_count, bin_sums_div_by_area


def get_features(segment_flow_mag, segment_flow_ang, mask, start, bin_size):
    '''
    '''
    
    feature_vector_collect = []
    
    for flow_mag, flow_ang in zip(segment_flow_mag, segment_flow_ang):
        
        # get the flow ang into radians
        flow_ang_radians = flow_ang[:, :, 0] * (360.0 / 255.0) * (np.pi / 180.0)
        
        bin_sums, bin_sums_div_by_count, bin_sums_div_by_area = \
            get_flow_mag_in_phase_bins(flow_mag[:, :, 0], flow_ang_radians,
                                       mask, start, bin_size)
            
        feature_vector_collect.append(bin_sums_div_by_count)
        
    feature_vector = np.hstack(feature_vector_collect).reshape((1, -1))
    
    return feature_vector
            
            
        
    
def function():

    directory = "./"
    
    for content in os.listdir(directory):
        
        if os.path.isfile(directory + content):
            
            if content[-12:] == '.sampled.mp4':
                
                name = content[:-12]
                
                break
                                       
    cap_flow_mag = cv2.VideoCapture(directory+'flow_'+name+'.flow_mag.mp4')
    cap_flow_ang = cv2.VideoCapture(directory+'flow_'+name+'.flow_ang.mp4')
    
    
    with open(directory + 'exemplars_and_labels___' + name + '.data',
              'rb') as f:
        
        exemplars_and_labels = pickle.load(f)
        
    exemplar_info = {}
    
    keys = np.array(list(exemplars_and_labels.keys()))
    
    
    segment = 45
    
    segment_flow_mag = []
    segment_flow_ang = []
    
    count = -1
    ind = -1
                
    num_bins = 8
    
    while(cap_flow_mag.isOpened() and cap_flow_ang.isOpened()):
        
        success_flag_flow_mag, flow_mag_display = cap_flow_mag.read()
        success_flag_flow_ang, flow_ang_display = cap_flow_ang.read()
        
        if success_flag_flow_mag == True and success_flag_flow_ang == True:
            
            count = count + 1
    
            segment_flow_mag.append(flow_mag_display)
            segment_flow_ang.append(flow_ang_display)
        
            if len(segment_flow_mag) > segment:
                segment_flow_mag.pop(0)
                segment_flow_ang.pop(0)
    
            else:
                ind = ind + 1
    
            if count != 0 and (count + 1) % segment == 0:
                
                print('segment', int(np.floor(count / segment)))
                
                index = np.where(keys[:, 1] == count * 4)[0]
                
                if len(index) == 1:
                    
                    exemplars, tags = \
                        exemplars_and_labels[(keys[index[0], 0],
                                              keys[index[0], 1])]
                        
                    feature_vectors = np.zeros((exemplars.shape[0],
                                                num_bins * segment))
                    
                    for i, box in enumerate(exemplars):
                        
                        col_1, row_1, col_2, row_2 = box
                        
                        mask = np.zeros(flow_mag_display.shape[0:2])
                        
                        mask[np.min([row_1, row_2]):np.max([row_1, row_2]),
                             np.min([col_1, col_2]):np.max([col_1, col_2])] = 1
                             
                        feature_vectors[i] = get_features(segment_flow_mag,
                            segment_flow_ang, mask, -2 * np.pi / num_bins / 2,
                            2 * np.pi / num_bins)
                        
                    exemplar_info[(keys[index[0], 0], keys[index[0], 1])] = \
                        [exemplars, tags, feature_vectors]
              
        else:
            
            print("No more frames")
            
            cap_flow_mag.release()
            cap_flow_ang.release()
                    
            break        
    else:
        
        cap_flow_mag.release()
        cap_flow_ang.release()
    
    with open(directory + \
        'exemplars_and_labels_and_feature_vectors_{:d}___'.format(num_bins) + \
        name + '.data', 'wb') as f:
        
        pickle.dump((exemplar_info), f)        