# -*- coding: utf-8 -*-
"""
Created on Mon May 14 18:37:17 2018

@author: Callie Darsey

subset_flow_mag_flow_ang_features_wheeler_01
"""



import cv2
import numpy as np
import pickle
import matplotlib.pyplot as plt
import os


def force_neg_pi_to_pi(value):
    '''
    This function takes the value and adds or subtracts pi until a value in
    [-pi, pi] is reached and returned.
    '''
    while value > np.pi or value < -np.pi:
        
        if value > np.pi:
            
            value = value - 2 * np.pi
            
        elif value < -np.pi:
            
            value = value + 2 * np.pi
            
    return value


    
def get_histogram_of_flow(flow_mag, flow_phase, start, bin_size):
    '''
    Return a histogram with flow phase bins [start, start + 1 * bin_size),
    [start + 2 * bin_size), ..., [start + n * size, start) adjusted such that
    the bins capture angles around the unit circle taking values [-pi, pi].
    
    For example, if start = -np.pi / 8 and bin_size = np.pi / 4, the bins are
    
    [    -pi / 8,      pi / 8),
    [     pi / 8,  3 * pi / 8),
    [ 3 * pi / 8,  5 * pi / 8),
    [ 5 * pi / 8,  7 * pi / 8),
    [ 7 * pi / 8, -7 * pi / 8),
    [-7 * pi / 8, -5 * pi / 8),
    [-5 * pi / 8, -3 * pi / 8),
    [-3 * pi / 8,     -pi / 8)
    
    Three versions of this histogram are returned.
    
    1) The value in each bin will be the sum of the flow magnitude values of
       pixels whose flow phase was within the phase bin range.
    
    2) The histogram described in 1) with each value divided by the phase bin
       count (if the count is nonzero)
       
    3) The histogram described in 1) with each value divided by the mask (i.e.
        component) area

    Inputs:
        
    flow_mag -  magnitude of the flow
    flow_phase - angle of the flow in radians
    mask - mask of the connected component
    start - angle in radians to start the phase bins at
    bin_size - angle in radians of the bin widths

    '''
    assert bin_size <= 2 * np.pi and bin_size > 0
    
    # force start to be in [-pi, pi]
    start = force_neg_pi_to_pi(start)
    
    # set up the bin edges
    bin_edges = [start]
    
    angles_travelled = 0
    while True:
         
        angles_travelled = angles_travelled + bin_size
        
        if angles_travelled >= 2 * np.pi:
            
            bin_edges.append(start)
            
            break
        
        else:

            new_edge = force_neg_pi_to_pi(bin_edges[-1] + bin_size)
            
            bin_edges.append(new_edge)

    # find the sums and counts for the bins
    
    bin_sums = np.zeros((1, len(bin_edges)-1), dtype=np.float32)
    bin_counts = np.zeros((1, len(bin_edges)-1), dtype=np.float32)
        
    for flow_mag_val, flow_phase_val in zip(flow_mag, flow_phase):
        
        flow_phase_value = force_neg_pi_to_pi(flow_phase_val)      
        
        for  i, edge_pair in enumerate(zip(bin_edges[0:-1], bin_edges[1:])):
            
            edge_inclusive, edge_exclusive = edge_pair
            
            # The bin does not cross pi on unit circle
            if edge_inclusive < edge_exclusive:
                
                if edge_inclusive <= flow_phase_value and \
                                     flow_phase_value < edge_exclusive:
                                       
                    bin_sums[0, i] = bin_sums[0, i] + flow_mag_val
                    
                    bin_counts[0, i] = bin_counts[0, i] + 1
                    
                    break
                    
            # The bin crosses pi on the unit circle
            elif edge_inclusive > edge_exclusive:
                
                if edge_inclusive <= flow_phase_value and \
                                     flow_phase_value <= np.pi or \
                           -np.pi <= flow_phase_value and \
                                     flow_phase_value < edge_exclusive:
                
                    bin_sums[0, i] = bin_sums[0, i] + flow_mag_val
                    
                    bin_counts[0, i] = bin_counts[0, i] + 1
                    
                    break
            
    return bin_sums, bin_counts



def combine(crop1, crop2):
    '''
    '''
    mask1 = np.any(crop1 > 50, axis=2)
    mask2 = np.any(crop2 > 50, axis=2)
    
    intersection = np.logical_and(mask1, mask2)
    xor_mask1 = np.logical_xor(mask1, intersection)
    xor_mask2 = np.logical_xor(mask2, intersection)    

    pred_side_1 = np.copy(crop1)
    pred_side_2 = np.copy(crop2)
    
    np.copyto(pred_side_1, crop2,
              where=np.stack([xor_mask2, xor_mask2, xor_mask2], axis=2))
    np.copyto(pred_side_2, crop1,
              where=np.stack([xor_mask1, xor_mask1, xor_mask1], axis=2))

    alpha = 0.5
    beta = 1.0 - alpha
    gamma = 0.0
    
    output = cv2.addWeighted(pred_side_1, alpha, pred_side_2, beta, gamma)
    
    return output


def get_seg_skin_overlays(segment_cleaned_skin, col_1, col_2, row_1, row_2):
    '''
    '''
    
    crop1 = segment_cleaned_skin[0][row_1 : row_2 + 1, col_1 : col_2 + 1, :]
    
    for index in np.arange(1, len(segment_cleaned_skin)):
        
        crop2 = segment_cleaned_skin[index][row_1 : row_2 + 1,
                                            col_1 : col_2 + 1, :]

        crop1 = combine(crop1, crop2)
        
    return crop1


def get_hists(segment_flow_mag, segment_flow_ang, row_1, row_2, col_1, col_2,
              bin_size):
    '''
    '''
    
    feature_flow_mag_hists = []
    
    feature_flow_ang_hists = []
    
    feature_hists_of_flow = []
    
    feature_hists_of_flow_count = []
    
    feature_hists_of_flow_div_counts = []
        
    for flow_mag, flow_ang in zip(segment_flow_mag, segment_flow_ang):
        
        flow_mag_patch = flow_mag[row_1 : row_2 + 1, col_1 : col_2 + 1].ravel()
        flow_ang_patch = flow_ang[row_1 : row_2 + 1, col_1 : col_2 + 1].ravel()
        
        feature_flow_mag_hists.append(np.histogram(flow_mag_patch,
                                                   bins=bin_size))
        
        feature_flow_ang_hists.append(np.histogram(flow_ang_patch,
                                                   bins=bin_size))
        
        bin_sums, bin_counts =  get_histogram_of_flow(flow_mag_patch,
            flow_ang_patch, -2 * np.pi / bin_size / 2, 2 * np.pi / bin_size)
        
        feature_hists_of_flow.append(bin_sums)
        
        feature_hists_of_flow_count.append(bin_counts)
        
        tmp = np.copy(bin_sums)
        
        tmp[bin_counts != 0] = \
            tmp[bin_counts != 0] / bin_counts[bin_counts != 0]
            
        feature_hists_of_flow_div_counts.append(tmp)


    feature_flow_mag_hists = np.vstack(feature_flow_mag_hists)
    
    feature_flow_ang_hists = np.vstack(feature_flow_ang_hists)
    
    feature_hists_of_flow = np.vstack(feature_hists_of_flow)
        
    feature_hists_of_flow_div_counts = \
        np.vstack(feature_hists_of_flow_div_counts)


    feature_flow_mag_hist = np.sum(feature_flow_mag_hists, axis=0)

    feature_flow_ang_hist = np.sum(feature_flow_ang_hists, axis=0)

    feature_hist_of_flow = np.sum(feature_hists_of_flow, axis=0)
    
    feature_sum_hists_of_flow_div_counts = \
        np.sum(feature_hists_of_flow_div_counts, axis=0)
        
    feature_hist_of_flow_div_counts = np.copy(feature_hist_of_flow)  
    feature_hist_of_flow_count = np.sum(np.vstack(feature_hists_of_flow_count),
                                        axis=0)

    feature_hist_of_flow_div_counts[feature_hist_of_flow_count != 0] = \
        feature_hist_of_flow_div_counts[feature_hist_of_flow_count != 0] / \
        feature_hist_of_flow_count[feature_hist_of_flow_count != 0]   

    
    return (feature_flow_mag_hists.ravel(), feature_flow_mag_hist,
            feature_flow_ang_hists.ravel(), feature_flow_ang_hist,
            feature_hists_of_flow.ravel(), 
            feature_hists_of_flow_div_counts.ravel(),
            feature_hist_of_flow, feature_sum_hists_of_flow_div_counts,
            feature_hist_of_flow_div_counts)




def get_subset_hists2(segment_flow_mag_display, segment_flow_ang_display,
                     row_1, row_2, col_1, col_2, bin_size):
    '''
    '''
    
    feature_flow_mag_hists = []
    
    feature_flow_ang_hists = []
    
    bins = np.arange(0, 256 + 256 / bin_size, 256 / bin_size)
           
    for flow_mag, flow_ang in zip(segment_flow_mag_display,
                                  segment_flow_ang_display):
        
        flow_mag_patch = flow_mag[row_1 : row_2 + 1, col_1 : col_2 + 1].ravel()
        flow_ang_patch = flow_ang[row_1 : row_2 + 1, col_1 : col_2 + 1].ravel()
        
        feature_flow_mag_hists.append(np.histogram(flow_mag_patch,
                                                   bins=bins)[0])
        
        feature_flow_ang_hists.append(np.histogram(flow_ang_patch,
                                                   bins=bins)[0])

    feature_flow_mag_hists = np.vstack(feature_flow_mag_hists)
    
    feature_flow_ang_hists = np.vstack(feature_flow_ang_hists)

    feature_flow_mag_hist = np.sum(feature_flow_mag_hists, axis=0)

    feature_flow_ang_hist = np.sum(feature_flow_ang_hists, axis=0)

    
    return (feature_flow_mag_hists.ravel(), feature_flow_mag_hist,
            feature_flow_ang_hists.ravel(), feature_flow_ang_hist)









def function(bin_size=32):
    '''
    '''
   
    e1 = cv2.getTickCount()
    
    directory = "./"
    
    for content in os.listdir(directory):
        
        if os.path.isfile(directory + content):
            
            if content[-12:] == '.sampled.mp4':
                
                name = content[:-12]
                
                print(name)
                
                break
                                       
    cap_flow_mag = cv2.VideoCapture(directory+'flow_'+name+'.flow_mag.mp4')
    cap_flow_ang = cv2.VideoCapture(directory+'flow_'+name+'.flow_ang.mp4')
    
    
    with open(directory + 'patches_info___' + name + '.data', 'rb') as f:
        
        patches_and_labels = pickle.load(f)
        
    keys = np.array(list(patches_and_labels.keys()))
    
    patches_and_labels_and_feature_flow_mag_hists = {}
    patches_and_labels_and_feature_flow_mag_hist = {}
    patches_and_labels_and_feature_flow_ang_hists = {}
    patches_and_labels_and_feature_flow_ang_hist = {}
    
    fdicts = [patches_and_labels_and_feature_flow_mag_hists,
              patches_and_labels_and_feature_flow_mag_hist,
              patches_and_labels_and_feature_flow_ang_hists,
              patches_and_labels_and_feature_flow_ang_hist]
    
    ftags = ['feature_flow_mag_hists',
             'feature_flow_mag_hist',
             'feature_flow_ang_hists',
             'feature_flow_ang_hist']
    
    
    

    
    
    segment = 45
    
    segment_flow_mag = []
    segment_flow_ang = []

    count = -1
    ind = -1
    
    while(cap_flow_mag.isOpened() and cap_flow_ang.isOpened()):
        
        success_flag_flow_mag, flow_mag_display    = cap_flow_mag.read(0)
        success_flag_flow_ang, flow_ang_display     = cap_flow_ang.read(0)
        
        if success_flag_flow_mag == True and success_flag_flow_ang == True:
            
            count = count + 1
            
            rows, cols, channels = flow_mag_display.shape
            
  
            segment_flow_mag.append(flow_mag_display[:, :, 0])
            segment_flow_ang.append(flow_ang_display[:, :, 0])
    
        
            if len(segment_flow_mag) > segment:
                segment_flow_mag.pop(0)
                segment_flow_ang.pop(0)

            else:
                ind = ind + 1
    
    
            if count != 0 and (count + 1) % segment == 0:
                
                ei = cv2.getTickCount()
                
                print('segment', int(np.floor(count / segment)))
    
                # find if segment has exemplars to collect features for
                index = np.where(keys[:, 1] == count * 4)[0]
    
                if len(index) == 1:
                    
                    # key for the segment
                    relevant_key = (keys[index[0], 0], keys[index[0], 1])
                    
                    # patch box info (and tags) for the segment
                    patches, tags = patches_and_labels[relevant_key]
    
                    # get the patch size
                    col_1, col_2 = np.sort([patches[0][0], patches[0][2]])
                
                    row_1, row_2 = np.sort([patches[0][1], patches[0][3]])
                    
                    patch_size = (row_2 - row_1 + 1, col_2 - col_1 + 1)
                    
    
                    '''operations on the  full frames over the segment'''
    
                    '''
                    ***********************
                    features to collect
                    ***********************
                    '''
                                        
                    ''' flow mag hist per patch per frame '''
                    feature_flow_mag_hists = np.zeros((patches.shape[0], bin_size * segment))
                    
                    ''' flow mag hist per patch through time '''                
                    feature_flow_mag_hist = np.zeros((patches.shape[0], bin_size))
    
                    ''' flow ang hist per patch per frame '''
                    feature_flow_ang_hists = np.zeros((patches.shape[0], bin_size * segment))
                    
                    ''' flow ang hist per patch through time '''
                    feature_flow_ang_hist = np.zeros((patches.shape[0], bin_size))
                    
                    
    
                    # loop through the patches
                    for i, patch in enumerate(patches):
                        
                        col_1, col_2 = np.sort([patch[0], patch[2]])
                    
                        row_1, row_2 = np.sort([patch[1], patch[3]])
    
                        feature_flow_mag_hists[i, :], \
                        feature_flow_mag_hist[i, :], \
                        feature_flow_ang_hists[i, :], \
                        feature_flow_ang_hist[i, :] = get_subset_hists2(\
                            segment_flow_mag, segment_flow_ang,
                            row_1, row_2, col_1, col_2, bin_size)
    
                    # put into the dictionaries
                    info_and_features = [feature_flow_mag_hists,
                                         feature_flow_mag_hist,
                                         feature_flow_ang_hists,
                                         feature_flow_ang_hist]                
                    
                    for info_or_feature, fdict in zip(info_and_features, fdicts):
                        
                        if (relevant_key) not in fdict.keys():
                            fdict[(relevant_key)] = []
                            fdict[(relevant_key)].append(patches)
                            fdict[(relevant_key)].append(tags)
                        
                        fdict[(relevant_key)].append(info_or_feature)
    
                eb = cv2.getTickCount()
                time = (eb - ei)/ cv2.getTickFrequency()
                print("Segment execution time in seconds = ", time)    
              
        else:
            
            print("No more frames")
            
            cap_flow_mag.release()
            cap_flow_ang.release()

            break        
    else:
        
        cap_flow_mag.release()
        cap_flow_ang.release()
    
    for ftag, fdict in zip(ftags, fdicts):
    
        with open(directory + 'patches_labels_and_subsubset_' + ftag + \
                  '_Bins_{:d}'.format(bin_size) + \
                  '___' + name + '.data', 'wb') as f:
        
            pickle.dump((fdict), f)
        
    e2 = cv2.getTickCount()
    time = (e2 - e1)/ cv2.getTickFrequency()
    print("Total execution time in seconds = ", time)
    
function(bin_size=32)